import os
from yaml import load, YAMLError
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper
import re
from redirector.utils.utils import get_map_path


class MapLineParser:
    def __init__(self, logger=None):
        self.url_regexp = re.compile(r'(\/[(\w)-:]+)+\/?(\.\w+)?')
        self.url_absolute = re.compile(
            r"^(?:http(s)?:\/\/)?[\w.-]+(?:\.[\w\.-]+)+[\w\-\._~:/?#[\]@!\$&'\(\)\*\+,;=.]+$")
        self.logger = logger

    def parse_line(self, line, prefix, i):
        line = line.strip()
        return_code, options = 0, []
        """
        return_codes:
        -1: blank line
        0 : default redirect
        1 : option redirect
        2 : absolute url
        3 : regexp
        """

        if line.startswith("#") or len(line) == 0:
            return -1, None, None

        line = line.split()

        if len(line) < 2:
            raise self.ParseLineError("Error on %s:{line};\n"
                                      "Not enough arguments to parse!".format(line=i), None)
        elif len(line) > 2:
            return_code = 1
            options = [option[1:-1] for option in line[2:]]
        print(line[0])

        if line[0].startswith("~"):
            line[0] = "~" + line[0][1:]            

        if line[0].startswith("//"):  # if new URI relative to the root
            line[0] = line[0][1:]  # cutting out extra '/' at the beginning
        elif line[1].startswith("//"):
            line[1] = line[1][1:]
            line[0] = line[0]

        elif self.url_absolute.fullmatch(line[1]):
            options = ['301']
            return_code = 1

        else:  # default url
            line[0] = line[0] if not line[0].startswith("~") else line[0]

        return return_code, line[0], line[1], options

    class ParseLineError(Exception):
        """Raised when line contains not enough arguments

        Attributes:
            message -- expanation where in .yaml config file aprser didn't find enough arguments
        """

        def __init__(self, message, errors):
            super().__init__(message)
            self.message = message
            self.errors = errors

    class RegexpTestError(Exception):
        """Raised when parser fails to compile url containing regular expression

        Attributes:
            message -- explanation what regexp failed to compile and where
        """

        def __init__(self, message, errors):
            super().__init__(message)
            self.message = message
            self.errors = errors


class ConfigReader:
    def __init__(self, logger=None):
        self.line_parser = MapLineParser(logger=logger)
        if logger:
            self.logger = logger

    @staticmethod
    def parse_yaml(file_dir):
        return_list = []
        yaml_dir = os.path.dirname(os.path.abspath(file_dir))

        stream = open(file_dir, 'r')

        data = load(stream, Loader=Loader)

        for project in data.get("projects"):
            map_path = project.get("map")
            
            abs_map_path = get_map_path(map_path, yaml_dir)

            project_prefix = project.get("prefix")
            return_list.append((abs_map_path, project_prefix))
        
        stream.close()

        return return_list

    def parse_map(self, map_file, yaml_file):
        res = [[], {}]
        res_prefix = None

        # normalize path     
        yaml_dir = os.path.dirname(os.path.abspath(yaml_file)) # get yaml file directory for restore rel path to map file
        map_dir = os.path.dirname(os.path.abspath(map_file))
        map_file_name = os.path.basename(map_file) # get filename from arg

        try:
            for map_path, prefix in self.parse_yaml(yaml_file):
                abs_map_path = get_map_path(map_path, map_dir)

                if map_file_name not in abs_map_path:
                    continue

                res_prefix = prefix.split("/")[-1]  #  Last directory of map_path a project's name

                f = open(abs_map_path, "r")

                for i, line in enumerate(f):
                    request_url, redirect_url, option = None, None, []
                    try:
                        return_code, request_url, redirect_url, *option = \
                                self.line_parser.parse_line(line, prefix, i)
                    except MapLineParser.RegexpTestError as e:
                        self.logger.log(e.message % map_file)
                    except MapLineParser.ParseLineError as e:
                        self.logger.log(e.message % map_file)

                    if request_url and redirect_url and return_code in [0, 1]:
                        if return_code == 0:
                            res[0].append((request_url, redirect_url))
                        elif return_code == 1:
                            opt_ = option[0][0] if option else option
                            
                            if opt_ in res[1].keys():
                                res[1][opt_].append([request_url, redirect_url])
                            else:
                                res[1][opt_] = [(request_url, redirect_url)]

                f.close()

            return res, res_prefix

        except YAMLError as e:
            self.logger.log("Error occurred while reading %s" % yaml_file + str(e))

